/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include "spamdyke.h"
#include "tls.h"
#include "dns.h"
#include "environment.h"
#include "command_line.h"
#include "usage.h"
#include "search_fs.h"
#include "exec.h"
#include "base64.h"

struct rejection_data *rejection = REJECTION_DATA;

void init_settings(struct filter_settings *target_settings)
  {
  target_settings->server_name[0] = '\0';
  target_settings->strlen_server_name = 0;
  target_settings->server_ip = NULL;
  target_settings->strlen_server_ip = 0;
  target_settings->local_server_name = NULL;

  target_settings->graylist_dir = NULL;
  target_settings->graylist_min_secs = 0;
  target_settings->graylist_max_secs = 0;
  target_settings->max_rcpt_to = 0;
  target_settings->use_syslog = 0;
  target_settings->policy_location = NULL;
  target_settings->strlen_policy_location = 0;
  target_settings->log_dir = NULL;
  target_settings->blacklist_sender_path = NULL;
  target_settings->blacklist_recipient_path = NULL;

  target_settings->access_list_path = NULL;
  target_settings->local_domains_path = NULL;
  target_settings->blacklist_rdns_keyword_path = NULL;
  target_settings->blacklist_rdns_directory_path = NULL;
  target_settings->blacklist_ip_path = NULL;
  target_settings->whitelist_rdns_path = NULL;
  target_settings->whitelist_ip_path = NULL;
  target_settings->dnsrbl_fqdn = NULL;
  target_settings->graylist_never_ip_path = NULL;
  target_settings->graylist_never_rdns_path = NULL;
  target_settings->graylist_always_ip_path = NULL;
  target_settings->graylist_always_rdns_path = NULL;
  target_settings->smtp_auth_command = NULL;
  target_settings->smtp_auth_encryption = 0;

  target_settings->graylist_type = GRAYLIST_NONE;
  target_settings->check_ip_in_rdns_cc = 0;
  target_settings->check_earlytalker = 0;
  target_settings->check_rdns_exist = 0;
  target_settings->check_rdns_resolve = 0;
  target_settings->check_sender_mx = 0;

  target_settings->timeout_connection = DEFAULT_TIMEOUT_CONNECTION_SECS;
  target_settings->timeout_command = DEFAULT_TIMEOUT_COMMAND_SECS;

  target_settings->allow_relay = 1;
  target_settings->inside_data = 0;
  target_settings->num_rcpt_to = 0;
  target_settings->local_sender = 0;

  target_settings->sender_address[0] = '\0';
  target_settings->recipient_address[0] = '\0';
  target_settings->graylist_path[0] = '\0';

  target_settings->rejection = NULL;
  target_settings->transient_rejection = NULL;
  target_settings->reject_message_buf[0] = '\0';

  target_settings->child_argv = NULL;
  target_settings->filter_action = FILTER_DECISION_UNDECIDED;

  target_settings->smtp_auth_type = SMTP_AUTH_UNKNOWN;
  target_settings->smtp_auth_origin = SMTP_AUTH_ORIGIN_NONE;
  target_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
  target_settings->smtp_auth_challenge[0] = '\0';
  target_settings->smtp_auth_response[0] = '\0';
  target_settings->smtp_auth_username[0] = '\0';

  target_settings->tls_state = TLS_STATE_INACTIVE;
#ifdef TLS
  target_settings->tls_context = NULL;
  target_settings->tls_session = NULL;
  target_settings->tls_certificate_path = NULL;
  target_settings->tls_privatekey_path = NULL;
  target_settings->tls_privatekey_password = NULL;
#endif // TLS

  return;
  }

int find_address(char *haystack, int strlen_haystack, char **return_needle, int *strlen_username)
  {
  int return_value;
  int found_username;
  char *target_address;

  return_value = 0;
  found_username = 0;

  if ((haystack != NULL) &&
      (strlen_haystack > 0) &&
      (return_needle != NULL))
    {
    for (target_address = haystack; (target_address < (haystack + strlen_haystack)) && (target_address[0] != ':'); target_address++);
    for (; (target_address < (haystack + strlen_haystack)) && !isalnum(target_address[0]); target_address++);

    for (return_value = 0; ((target_address + return_value) < (haystack + strlen_haystack)); return_value++)
      if (!isalnum(target_address[return_value]))
        {
        switch (target_address[return_value])
          {
          case '@':
            if (strlen_username != NULL)
              *strlen_username = return_value;

            found_username = 1;

          case ':':
          case ',':
          case '.':
          case '#':
          case '[':
          case ']':
          case '-':
          case '"':
          case '\\':
          case '_':
            continue;
            break;
          }

        break;
        }

    *return_needle = (return_value > 0) ? target_address : NULL;
    if ((strlen_username != NULL) &&
        !found_username)
      *strlen_username = return_value;
    }

  return(return_value);
  }

int log_writeln(struct filter_settings *current_settings, int action, int target_fd, char *data, int data_length)
  {
  static FILE *log_file = NULL;
  static int last_action = -1;
  static char *prefixes[] = LOG_ACTION_PREFIX;
  char log_filename[MAX_BUF + 1];
  struct tm *tmp_tm;
  time_t tmp_time;

  if (action != LOG_ACTION_NONE)
    {
    tmp_tm = NULL;

    if ((current_settings == NULL) &&
        (log_file != NULL))
      {
      tmp_time = time(NULL);
      tmp_tm = localtime(&tmp_time);

      fprintf(log_file, "\n%.2d/%.2d/%d %.2d:%.2d:%.2d CLOSED\n", tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

      fclose(log_file);
      log_file = NULL;
      }
    else if ((current_settings != NULL) &&
             (current_settings->log_dir != NULL) &&
             (log_file == NULL))
      {
      tmp_time = time(NULL);
      tmp_tm = localtime(&tmp_time);

      if (current_settings->strlen_server_name > 0)
        snprintf(log_filename, MAX_BUF, "%s/%d%.2d%.2d_%.2d%.2d%.2d_%s_%s", current_settings->log_dir, tmp_tm->tm_year + 1900, tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, current_settings->server_ip, current_settings->server_name);
      else
        snprintf(log_filename, MAX_BUF, "%s/%d%.2d%.2d_%.2d%.2d%.2d_%s", current_settings->log_dir, tmp_tm->tm_year + 1900, tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, current_settings->server_ip);

      if ((log_file = fopen(log_filename, "w")) != NULL)
        {
        chmod(log_filename, CHMOD_MODE);
        fprintf(log_file, "%.2d/%.2d/%d %.2d:%.2d:%.2d STARTED: To remote host = %s; to child process = %s\n", tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec, prefixes[LOG_ACTION_CHILD_FROM], prefixes[LOG_ACTION_REMOTE_FROM]);
        }
      else
        {
        syslog(LOG_ERR, SYSLOG_ERROR_OPEN_LOG "%s: %s", log_filename, strerror(errno));
        current_settings->log_dir = NULL;
        }
      }

    if (log_file != NULL)
      {
      if (tmp_tm == NULL)
        {
        tmp_time = time(NULL);
        tmp_tm = localtime(&tmp_time);
        }

      switch (action)
        {
        case LOG_ACTION_TLS_PASSTHROUGH_START:
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->smtp_auth_username[0] != '\0')
            fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", current_settings->smtp_auth_username);

          fprintf(log_file, " - %s\n", LOG_MESSAGE_TLS_PASSTHROUGH_START);
          break;
        case LOG_ACTION_TLS_START:
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->smtp_auth_username[0] != '\0')
            fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", current_settings->smtp_auth_username);

          fprintf(log_file, " - %s\n", LOG_MESSAGE_TLS_START);
          break;
        case LOG_ACTION_TLS_END:
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->smtp_auth_username[0] != '\0')
            fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", current_settings->smtp_auth_username);

          fprintf(log_file, " - %s\n", LOG_MESSAGE_TLS_END);
          break;
        case LOG_ACTION_AUTH_SUCCESS:
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
            fprintf(log_file, LOG_ACTION_PREFIX_TLS_SPAMDYKE);
          else if (current_settings->tls_state == TLS_STATE_ACTIVE_PASSTHROUGH)
            fprintf(log_file, LOG_ACTION_PREFIX_TLS_PASSTHROUGH);

          if (current_settings->smtp_auth_username[0] != '\0')
            fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", current_settings->smtp_auth_username);

          fprintf(log_file, " - %s\n", LOG_MESSAGE_AUTH_SUCCESS);
          break;
        case LOG_ACTION_AUTH_FAILURE:
          fprintf(log_file, "\n%s", prefixes[LOG_ACTION_FILTER_FROM]);
          fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[LOG_ACTION_FILTER_FROM], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

          if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
            fprintf(log_file, LOG_ACTION_PREFIX_TLS_SPAMDYKE);
          else if (current_settings->tls_state == TLS_STATE_ACTIVE_PASSTHROUGH)
            fprintf(log_file, LOG_ACTION_PREFIX_TLS_PASSTHROUGH);

          fprintf(log_file, " - %s%s\n", LOG_MESSAGE_AUTH_FAILURE, data);
          break;
        default:
          if (action != last_action)
            {
            fprintf(log_file, "\n%s %.2d/%.2d/%d %.2d:%.2d:%.2d", prefixes[action], tmp_tm->tm_mon + 1, tmp_tm->tm_mday, tmp_tm->tm_year + 1900, tmp_tm->tm_hour, tmp_tm->tm_min, tmp_tm->tm_sec);

            if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
              fprintf(log_file, LOG_ACTION_PREFIX_TLS_SPAMDYKE);
            else if (current_settings->tls_state == TLS_STATE_ACTIVE_PASSTHROUGH)
              fprintf(log_file, LOG_ACTION_PREFIX_TLS_PASSTHROUGH);

            if (current_settings->smtp_auth_username[0] != '\0')
              fprintf(log_file, LOG_ACTION_PREFIX_AUTH "%s", current_settings->smtp_auth_username);

            fprintf(log_file, "\n");
            }
          break;
        }

      last_action = action;

      if (data_length >= 0)
        {
        if (current_settings->tls_state == TLS_STATE_ACTIVE_PASSTHROUGH)
          {
          int i;

          for (i = 0; i < data_length; i++)
            fprintf(log_file, "%.2x ", data[i]);
          }
        else
          fprintf(log_file, "%.*s", data_length, data);
        }

      fflush(NULL);
      }
    }

  return((target_fd >= 0) ? (((action == LOG_ACTION_CHILD_FROM) || (action == LOG_ACTION_FILTER_FROM)) ? NETWORK_WRITE(current_settings, target_fd, data, data_length) : write(target_fd, data, data_length)) : 0);
  }

int log_write_rejection(struct filter_settings *current_settings, struct rejection_data *target_rejection, int target_fd, char *substitute_reject_severity)
  {
  int return_value;

  if ((return_value = (substitute_reject_severity == NULL) ? log_writeln(current_settings, LOG_ACTION_FILTER_FROM, target_fd, target_rejection->reject_severity, STRLEN_REJECT_SEVERITY) : log_writeln(current_settings, LOG_ACTION_FILTER_FROM, target_fd, substitute_reject_severity, STRLEN_REJECT_SEVERITY)) != 0)
    return_value = log_writeln(current_settings, LOG_ACTION_FILTER_FROM, target_fd, target_rejection->reject_message, target_rejection->strlen_reject_message);

  if (return_value &&
      (current_settings->policy_location != NULL))
    {
    if ((return_value = log_writeln(current_settings, LOG_ACTION_FILTER_FROM, target_fd, ERROR_URL, STRLEN(ERROR_URL))) != 0)
      return_value = log_writeln(current_settings, LOG_ACTION_FILTER_FROM, target_fd, current_settings->policy_location, current_settings->strlen_policy_location);
    }

  if (return_value)
    return_value = log_writeln(current_settings, LOG_ACTION_FILTER_FROM, target_fd, REJECT_CRLF, STRLEN(REJECT_CRLF));

  fflush(NULL);

  return(return_value);
  }

int smtp_filter(int inbound_fd, int outbound_fd, char *input_line, int strlen_input_line, struct filter_settings *current_settings, char *envp[])
  {

#ifdef TLS
  static struct rejection_data tls_success = SUCCESS_TLS;
#endif /* TLS */

  static struct rejection_data tls_failure = FAILURE_TLS;
  static struct rejection_data smtp_auth_success = SUCCESS_AUTH;
  static char *smtp_auth_types[] = SMTP_AUTH_TYPES;
  static int strlen_smtp_auth_types[] = STRLEN_SMTP_AUTH_TYPES;
  static int smtp_auth_encryption[] = SMTP_AUTH_ENCRYPTION;
  int return_value;
  int i;
  char *target_address;
  int strlen_target_address;
  int strlen_username;
  struct stat tmp_stat;
  FILE *tmp_file;
  char log_entry[MAX_BUF + 1];
  int continue_processing;
  char tmp_recipient_domain[MAX_BUF + 1];
  char tmp_recipient_username[MAX_BUF + 1];
  char tmp_sender_address[MAX_BUF + 1];
  int search_results;
  char decoded_content[MAX_BUF + 1];
  int strlen_content;
  char username[MAX_BUF + 1];
  char password[MAX_BUF + 1];
  int local_recipient;

  return_value = FILTER_FLAG_PASS;

  if (current_settings->filter_action == FILTER_DECISION_PASS_TLS)
    return_value = FILTER_FLAG_PASS;
  else if (current_settings->inside_data)
    {
    if ((strlen_input_line == STRLEN(SMTP_DATA_END)) &&
        (strncasecmp(SMTP_DATA_END, input_line, STRLEN(SMTP_DATA_END)) == 0))
      current_settings->inside_data = 0;
    }
  else if ((current_settings->filter_action == FILTER_DECISION_CHECK_AUTH) &&
           (current_settings->smtp_auth_state == SMTP_AUTH_STATE_CHALLENGE_1_SENT))
    {
    current_settings->smtp_auth_state = SMTP_AUTH_STATE_RESPONSE_1_SEEN;

    switch (current_settings->smtp_auth_type)
      {
      case SMTP_AUTH_PLAIN:
        strlen_content = base64_decode((unsigned char *)decoded_content, MAX_BUF, (unsigned char *)input_line, strlen_input_line);
        decoded_content[strlen_content] = '\0';
        if ((strlen_content >= 3) &&
            (decoded_content[0] == '\0'))
          {
          snprintf(username, MAX_BUF, "%s", decoded_content + 1);
          for (i = 1; ((i < strlen_content) && (decoded_content[i] != '\0')); i++);
          if (i < strlen_content)
            {
            snprintf(password, MAX_BUF, "%s", decoded_content + i + 1);

            for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
              if (exec_checkpassword(current_settings->smtp_auth_command[i], username, password, NULL, envp))
                {
                if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
                  {
                  log_write_rejection(current_settings, &smtp_auth_success, outbound_fd, NULL);
                  return_value = FILTER_FLAG_INTERCEPT;
                  }
                else
                  return_value = FILTER_FLAG_PASS;

                snprintf(current_settings->smtp_auth_username, MAX_BUF, "%s", username);
                current_settings->filter_action = FILTER_DECISION_BYPASS_FILTER;
                current_settings->rejection = NULL;

                log_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, NULL, -1);
                break;
                }
            }
          }

        if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
          {
          log_writeln(current_settings, LOG_ACTION_AUTH_FAILURE, -1, username, -1);

          if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
            {
            log_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
            return_value = FILTER_FLAG_INTERCEPT | ((current_settings->rejection == NULL) ? FILTER_FLAG_CHILD_CONTINUE : FILTER_FLAG_CHILD_QUIT);
            }
          else if (current_settings->rejection != NULL)
            {
            log_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
            return_value = (FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT);
            }
          else
            return_value = FILTER_FLAG_PASS;

          current_settings->filter_action = FILTER_DECISION_DO_FILTER;
          }

        current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;

        break;
      case SMTP_AUTH_LOGIN:
        strncpy(current_settings->smtp_auth_response, input_line, MINVAL(MAX_BUF, strlen_input_line));
        current_settings->smtp_auth_response[MINVAL(MAX_BUF, strlen_input_line)] = '\0';

        if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
          {
          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_AUTH_CHALLENGE, STRLEN_REJECT_SEVERITY);
          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, SMTP_AUTH_LOGIN_CHALLENGE_2, STRLEN(SMTP_AUTH_LOGIN_CHALLENGE_2));
          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN(REJECT_CRLF));
          current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_2_SENT;

          return_value = FILTER_FLAG_INTERCEPT;
          }
        else
          return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;

        break;
      case SMTP_AUTH_CRAM_MD5:
        strlen_content = base64_decode((unsigned char *)decoded_content, MAX_BUF, (unsigned char *)input_line, strlen_input_line);
        decoded_content[strlen_content] = '\0';

        if (strlen_content > 0)
          {
          for (i = 0; (i < strlen_content) && (decoded_content[i] != ' '); i++);

          if (i < strlen_content)
            {
            snprintf(username, MAX_BUF, "%.*s", i, decoded_content);
            snprintf(password, MAX_BUF, "%.*s", (strlen_content - i) - 1, decoded_content + i + 1);

            strlen_content = strlen(current_settings->smtp_auth_challenge);
            for (i = STRLEN(SMTP_AUTH_CHALLENGE); (i < strlen_content) && (current_settings->smtp_auth_challenge[i] != ' '); i++);
            if (i < strlen_content)
              {
              strlen_content = base64_decode((unsigned char *)decoded_content, MAX_BUF, (unsigned char *)current_settings->smtp_auth_challenge + i, strlen_content - i);
              decoded_content[strlen_content] = '\0';

              for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
                if (exec_checkpassword(current_settings->smtp_auth_command[i], username, password, decoded_content, envp))
                  {
                  if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
                    {
                    log_write_rejection(current_settings, &smtp_auth_success, outbound_fd, NULL);
                    return_value = FILTER_FLAG_INTERCEPT;
                    }
                  else
                    return_value = FILTER_FLAG_PASS;

                  snprintf(current_settings->smtp_auth_username, MAX_BUF, "%s", username);
                  current_settings->filter_action = FILTER_DECISION_BYPASS_FILTER;
                  current_settings->rejection = NULL;

                  log_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, NULL, -1);
                  break;
                  }
              }
            }
          }

        if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
          {
          log_writeln(current_settings, LOG_ACTION_AUTH_FAILURE, -1, username, -1);

          if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
            {
            log_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
            return_value = FILTER_FLAG_INTERCEPT | ((current_settings->rejection == NULL) ? FILTER_FLAG_CHILD_CONTINUE : FILTER_FLAG_CHILD_QUIT);
            }
          else if (current_settings->rejection != NULL)
            {
            log_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
            return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
            }
          else
            return_value = FILTER_FLAG_PASS;

          current_settings->filter_action = FILTER_DECISION_DO_FILTER;
          }

        current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;

        break;
      default:
        // The client is using a protocol spamdyke doesn't support. Continue listening for the server's response and trust it.
        current_settings->smtp_auth_state = SMTP_AUTH_STATE_RESPONSE_2_SEEN;
        return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
        break;
      }
    }
  else if ((current_settings->filter_action == FILTER_DECISION_CHECK_AUTH) &&
           (current_settings->smtp_auth_state == SMTP_AUTH_STATE_CHALLENGE_2_SENT))
    {
    switch (current_settings->smtp_auth_type)
      {
      case SMTP_AUTH_LOGIN:
        strlen_content = base64_decode((unsigned char *)username, MAX_BUF, (unsigned char *)current_settings->smtp_auth_response, strlen(current_settings->smtp_auth_response));
        username[strlen_content] = '\0';
        strlen_content = base64_decode((unsigned char *)password, MAX_BUF, (unsigned char *)input_line, strlen_input_line);
        password[strlen_content] = '\0';

        for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
          if (exec_checkpassword(current_settings->smtp_auth_command[i], username, password, NULL, envp))
            {
            if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
              {
              log_write_rejection(current_settings, &smtp_auth_success, outbound_fd, NULL);
              return_value = FILTER_FLAG_INTERCEPT;
              }
            else
              return_value = FILTER_FLAG_PASS;

            snprintf(current_settings->smtp_auth_username, MAX_BUF, "%s", username);
            current_settings->filter_action = FILTER_DECISION_BYPASS_FILTER;
            current_settings->rejection = NULL;

            log_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, NULL, -1);
            break;
            }

        if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
          {
          log_writeln(current_settings, LOG_ACTION_AUTH_SUCCESS, -1, username, -1);

          if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
            {
            log_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
            return_value = FILTER_FLAG_INTERCEPT | ((current_settings->rejection == NULL) ? FILTER_FLAG_CHILD_CONTINUE : FILTER_FLAG_CHILD_QUIT);
            }
          else if (current_settings->rejection != NULL)
            {
            log_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_FAILURE], outbound_fd, NULL);
            return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
            }
          else
            return_value = FILTER_FLAG_PASS;

          current_settings->filter_action = FILTER_DECISION_DO_FILTER;
          }

        current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;

        break;
      default:
        // The client is using a protocol spamdyke doesn't support. Continue listening for the server's response and trust it.
        current_settings->smtp_auth_state = SMTP_AUTH_STATE_RESPONSE_2_SEEN;
        return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
        break;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_HELO)) &&
           (strncasecmp(SMTP_HELO, input_line, STRLEN(SMTP_HELO)) == 0))
    {
    if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
      current_settings->filter_action = FILTER_DECISION_DO_FILTER;

    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->rejection != NULL))
      {
      log_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_NONE);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_EHLO)) &&
           (strncasecmp(SMTP_EHLO, input_line, STRLEN(SMTP_EHLO)) == 0))
    {
    if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
      {
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_ADD;

#ifdef TLS
      if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
        return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_ADD | FILTER_FLAG_TLS_REMOVE;
      else if (current_settings->tls_certificate_path != NULL)
        return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_ADD | FILTER_FLAG_TLS_ADD;
      }
    else if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_TLS_REMOVE;
    else if (current_settings->tls_certificate_path != NULL)
      {
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_TLS_ADD;
#endif /* TLS */

      }
    else if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
             (current_settings->rejection != NULL))
      {
      log_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_NONE);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }
  else if ((current_settings->filter_action == FILTER_DECISION_CHECK_AUTH) &&
           (strlen_input_line >= STRLEN(SMTP_AUTH)) &&
           (strncasecmp(SMTP_AUTH, input_line, STRLEN(SMTP_AUTH)) == 0))
    {
    for (i = 0; smtp_auth_types[i] != NULL; i++)
      if ((strlen_input_line >= (strlen_smtp_auth_types[i] + STRLEN(SMTP_AUTH) + 1)) &&
          (strncasecmp(input_line + STRLEN(SMTP_AUTH) + 1, smtp_auth_types[i], strlen_smtp_auth_types[i]) == 0) &&
          (current_settings->smtp_auth_encryption ||
           !smtp_auth_encryption[i]))
        {
        current_settings->smtp_auth_type = i;
        break;
        }

    current_settings->smtp_auth_state = SMTP_AUTH_STATE_CMD_SEEN;

    if (current_settings->smtp_auth_origin == SMTP_AUTH_ORIGIN_SPAMDYKE)
      {
      switch (current_settings->smtp_auth_type)
        {
        case SMTP_AUTH_PLAIN:
          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_AUTH_CHALLENGE, STRLEN_REJECT_SEVERITY);
          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN(REJECT_CRLF));
          current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_1_SENT;
          break;
        case SMTP_AUTH_LOGIN:
          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_SEVERITY_AUTH_CHALLENGE, STRLEN_REJECT_SEVERITY);
          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, SMTP_AUTH_LOGIN_CHALLENGE_1, STRLEN(SMTP_AUTH_LOGIN_CHALLENGE_1));
          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN(REJECT_CRLF));
          current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_1_SENT;
          break;
        case SMTP_AUTH_CRAM_MD5:
          strlen_content = snprintf(decoded_content, MAX_BUF, "<%ld.%ld@%s>", random(), (long)time(NULL), (current_settings->local_server_name != NULL) ? current_settings->local_server_name : MISSING_LOCAL_SERVER_NAME);
          snprintf(current_settings->smtp_auth_challenge, MAX_BUF, "%s ", SMTP_AUTH_CHALLENGE);

          strlen_content = base64_encode((unsigned char *)(current_settings->smtp_auth_challenge + STRLEN(SMTP_AUTH_CHALLENGE) + 1), MAX_BUF - (STRLEN(SMTP_AUTH_CHALLENGE) + 1), (unsigned char *)decoded_content, strlen_content) + STRLEN(SMTP_AUTH_CHALLENGE) + 1;
          current_settings->smtp_auth_challenge[strlen_content] = '\0';

          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, current_settings->smtp_auth_challenge, strlen_content);
          log_writeln(current_settings, LOG_ACTION_FILTER_FROM, outbound_fd, REJECT_CRLF, STRLEN(REJECT_CRLF));

          current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_1_SENT;

          break;
        default:
          // The client requested an algorithm spamdyke didn't advertise.
          log_write_rejection(current_settings, &rejection[REJECTION_SMTP_AUTH_UNKNOWN], outbound_fd, NULL);
          current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
          current_settings->filter_action = FILTER_DECISION_DO_FILTER;
          break;
        }

      return_value = FILTER_FLAG_INTERCEPT;
      }
    else
      return_value = FILTER_FLAG_PASS | FILTER_FLAG_AUTH_CAPTURE;
    }
  else if ((strlen_input_line >= STRLEN(SMTP_TLS)) &&
           (strncasecmp(SMTP_TLS, input_line, STRLEN(SMTP_TLS)) == 0))
    {
    if (current_settings->tls_state == TLS_STATE_INACTIVE)
      if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
          (current_settings->rejection != NULL))
        {
        log_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_NONE);
        return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
        }
      else

#ifdef TLS
        if (current_settings->tls_certificate_path != NULL)
          {
          log_write_rejection(current_settings, &tls_success, outbound_fd, NULL);
          if (tls_start(current_settings, inbound_fd, outbound_fd))
            log_writeln(current_settings, LOG_ACTION_TLS_START, -1, NULL, -1);
          else
            log_write_rejection(current_settings, &tls_failure, outbound_fd, NULL);

          return_value = FILTER_FLAG_INTERCEPT;
          }
        else
#endif /* TLS */

          return_value = FILTER_FLAG_PASS | FILTER_FLAG_TLS_CAPTURE;
    else
      {
      log_write_rejection(current_settings, &tls_failure, outbound_fd, REJECT_SEVERITY_PERMANENT);
      return_value = FILTER_FLAG_INTERCEPT;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_MAIL_FROM)) &&
           (strncasecmp(SMTP_MAIL_FROM, input_line, STRLEN(SMTP_MAIL_FROM)) == 0))
    {
    if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
      current_settings->filter_action = FILTER_DECISION_DO_FILTER;

    if (((strlen_target_address = find_address(input_line, strlen_input_line, &target_address, &strlen_username)) > 0) &&
        (strlen_target_address < MAX_ADDRESS))
      {
      for (i = 0; i < strlen_target_address; i++)
        current_settings->sender_address[i] = tolower(target_address[i]);
      current_settings->sender_address[strlen_target_address] = '\0';

      if ((current_settings->blacklist_sender_path == NULL) ||
          (search_file(current_settings->blacklist_sender_path, current_settings->sender_address, strlen_target_address, '@', '\0') != 1))
        {
        if (!current_settings->local_sender &&
            (current_settings->local_domains_path != NULL))
          for (i = 0; current_settings->local_domains_path[i] != NULL; i++)
            if ((current_settings->local_sender = search_file(current_settings->local_domains_path[i], target_address + strlen_username + 1, (strlen_target_address - strlen_username) - 1, '.', '\0')) == 1)
              break;

        if ((current_settings->rejection == NULL) &&
            !current_settings->local_sender &&
            current_settings->check_sender_mx &&
            (((strlen_target_address - strlen_username) - 1) > 0) &&
            !dns_mx(target_address + strlen_username + 1, (strlen_target_address - strlen_username) - 1))
          current_settings->rejection = &rejection[REJECTION_SENDER_NO_MX];
        }
      else if (current_settings->rejection == NULL)
        current_settings->rejection = &rejection[REJECTION_SENDER_BLACKLISTED];
      }

    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->rejection != NULL))
      {
      log_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_NONE);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_RCPT_TO)) &&
           (strncasecmp(SMTP_RCPT_TO, input_line, STRLEN(SMTP_RCPT_TO)) == 0))
    {
    if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
      current_settings->filter_action = FILTER_DECISION_DO_FILTER;

    if ((strlen_target_address = find_address(input_line, strlen_input_line, &target_address, &strlen_username)) > 0)
      {
      for (i = 0; i < strlen_target_address; i++)
        current_settings->recipient_address[i] = tolower(target_address[i]);
      current_settings->recipient_address[strlen_target_address] = '\0';

      if (current_settings->local_domains_path != NULL)
        {
        local_recipient = 0;
        for (i = 0; current_settings->local_domains_path[i] != NULL; i++)
          if ((local_recipient = search_file(current_settings->local_domains_path[i], current_settings->recipient_address + strlen_username + 1, (strlen_target_address - strlen_username) - 1, '.', '\0')) == 1)
            break;
        }
      else
        local_recipient = 1;
      }
    else
      local_recipient = 1;

    if (current_settings->filter_action == FILTER_DECISION_DO_FILTER)
      {
      if (current_settings->rejection != NULL)
        {
        log_write_rejection(current_settings, current_settings->rejection, outbound_fd, NULL);
        return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
        }
      else if (strlen_username == strlen_target_address)
        {
        current_settings->transient_rejection = &rejection[REJECTION_RCPT_TO_LOCAL];
        log_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);
        return_value = FILTER_FLAG_INTERCEPT;
        }
      else if (!local_recipient &&
               !current_settings->allow_relay)
        {
        current_settings->transient_rejection = &rejection[REJECTION_RELAYING_DENIED];
        log_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);
        return_value = FILTER_FLAG_INTERCEPT;
        }
      else if ((current_settings->max_rcpt_to > 0) &&
               (current_settings->num_rcpt_to >= current_settings->max_rcpt_to))
        {
        current_settings->transient_rejection = &rejection[REJECTION_RCPT_TO];
        log_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);
        return_value = FILTER_FLAG_INTERCEPT;
        }
      else if ((current_settings->blacklist_recipient_path != NULL) &&
               (search_file(current_settings->blacklist_recipient_path, current_settings->recipient_address, strlen_target_address, '@', '@') == 1))
        {
        current_settings->transient_rejection = &rejection[REJECTION_RECIPIENT_BLACKLISTED];
        log_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);
        return_value = FILTER_FLAG_INTERCEPT;
        }
      else if (current_settings->graylist_type != GRAYLIST_NONE)
        {
        // Figure out what kind of graylisting is wanted and if it's needed for this connection.
        if (current_settings->graylist_type == GRAYLIST_FILTER_NONE)
          {
          continue_processing = 0;

          if (current_settings->graylist_always_ip_path != NULL)
            for (i = 0; current_settings->graylist_always_ip_path[i] != NULL; i++)
              if (search_tcprules_file(NULL, 0, current_settings->graylist_always_ip_path[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name) == 1)
                {
                continue_processing = 1;
                break;
                }

          if (!continue_processing &&
              (current_settings->graylist_always_rdns_path != NULL))
            for (i = 0; current_settings->graylist_always_rdns_path[i] != NULL; i++)
              if (search_file(current_settings->graylist_always_rdns_path[i], current_settings->server_name, current_settings->strlen_server_name, '.', '\0') == 1)
                {
                continue_processing = 1;
                break;
                }
          }
        else
          {
          continue_processing = 1;

          if (current_settings->graylist_never_ip_path != NULL)
            for (i = 0; current_settings->graylist_never_ip_path[i] != NULL; i++)
              if (search_tcprules_file(NULL, 0, current_settings->graylist_never_ip_path[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name) == 1)
                {
                continue_processing = 0;
                break;
                }

          if (continue_processing &&
              (current_settings->graylist_never_rdns_path != NULL))
            for (i = 0; current_settings->graylist_never_rdns_path[i] != NULL; i++)
              if (search_file(current_settings->graylist_never_rdns_path[i], current_settings->server_name, current_settings->strlen_server_name, '.', '\0') == 1)
                {
                continue_processing = 0;
                break;
                }
          }

        // Construct the path to the recipient's domain's graylist folder and check if it exists
        if (continue_processing &&
            (strlen_target_address > 0) &&
            (snprintf(current_settings->graylist_path, MAX_PATH, "%s/%s", current_settings->graylist_dir, canonicalize_path(tmp_recipient_domain, MAX_PATH, target_address + strlen_username + 1, (strlen_target_address - strlen_username) - 1)) < MAX_PATH) &&
            (stat(current_settings->graylist_path, &tmp_stat) == 0) &&
            (S_ISDIR(tmp_stat.st_mode)))
          {
          continue_processing = 1;
          if ((snprintf(current_settings->graylist_path, MAX_PATH, "%s/%s/%s", current_settings->graylist_dir, tmp_recipient_domain, canonicalize_path(tmp_recipient_username, MAX_PATH, target_address, strlen_username)) < MAX_PATH) &&
              (stat(current_settings->graylist_path, &tmp_stat) != 0))
            {
            if (errno == ENOENT)
              {
              if (mkdir(current_settings->graylist_path, MKDIR_MODE) != 0)
                {
                continue_processing = 0;
                syslog(LOG_ERR, SYSLOG_ERROR_MKDIR "%s: %s", current_settings->graylist_path, strerror(errno));
                }
              }
            else
              {
              continue_processing = 0;
              syslog(LOG_ERR, SYSLOG_ERROR_STAT "%s: %s", current_settings->graylist_path, strerror(errno));
              }
            }

          // Construct the path to the sender->user file and check if it exists
          if (continue_processing &&
              (snprintf(current_settings->graylist_path, MAX_PATH, "%s/%s/%s/%s", current_settings->graylist_dir, tmp_recipient_domain, tmp_recipient_username, (current_settings->sender_address[0] != '\0') ? canonicalize_path(tmp_sender_address, MAX_PATH, current_settings->sender_address, strlen(current_settings->sender_address)) : SENDER_ADDRESS_NONE) < MAX_PATH) &&
              (stat(current_settings->graylist_path, &tmp_stat) == 0))
            // Check that the file is a "regular" file and the age matches the command line options
            if ((S_ISREG(tmp_stat.st_mode)) &&
                (((current_settings->graylist_min_secs > 0) &&
                  (((tmp_stat.st_mtime + current_settings->graylist_min_secs) <= time(NULL)) ||
                   (tmp_stat.st_size > 0))) ||
                 (current_settings->graylist_min_secs == 0)) &&
                (((current_settings->graylist_max_secs > 0) &&
                  ((tmp_stat.st_mtime + current_settings->graylist_max_secs) >= time(NULL))) ||
                 (current_settings->graylist_max_secs == 0)))
              {
              // User was graylisted but is now allowed.  Log the name of their sending host to the file.
              if (current_settings->strlen_server_name > 0)
                snprintf(log_entry, MAX_BUF, "%s %s", current_settings->server_ip, current_settings->server_name);
              else
                snprintf(log_entry, MAX_BUF, "%s", current_settings->server_ip);

              search_results = search_file(current_settings->graylist_path, log_entry, strlen(log_entry), '\0', '\0');
              if ((tmp_file = fopen(current_settings->graylist_path, "a")) != NULL)
                {
                chmod(current_settings->graylist_path, CHMOD_MODE);

                if (search_results != 1)
                  fprintf(tmp_file, "%s\n", log_entry);

                fclose(tmp_file);
                }
              else
                syslog(LOG_ERR, SYSLOG_ERROR_GRAYLIST_FILE "%s: %s", current_settings->graylist_path, strerror(errno));
              }
            else
              {
              // The file's age is outside the bounds of the command line options.
              return_value = FILTER_FLAG_INTERCEPT;

              // The file was too old -- truncate it.
              if ((current_settings->graylist_max_secs > 0) &&
                  ((tmp_stat.st_mtime + current_settings->graylist_max_secs) <= time(NULL)))
                {
                if ((tmp_file = fopen(current_settings->graylist_path, "w")) != NULL)
                  {
                  chmod(current_settings->graylist_path, CHMOD_MODE);

                  fclose(tmp_file);
                  }
                else
                  {
                  return_value = FILTER_FLAG_PASS;
                  syslog(LOG_ERR, SYSLOG_ERROR_GRAYLIST_FILE "%s: %s", current_settings->graylist_path, strerror(errno));
                  }
                }

              if ((return_value & FILTER_MASK_PASS) != FILTER_FLAG_PASS)
                {
                current_settings->transient_rejection = &rejection[REJECTION_GRAYLISTED];
                log_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);
                }
              }
          // Sender->user graylist file does not exist -- create it.
          else if (continue_processing &&
                   ((tmp_file = fopen(current_settings->graylist_path, "w")) != NULL))
            {
            chmod(current_settings->graylist_path, CHMOD_MODE);

            fclose(tmp_file);

            current_settings->transient_rejection = &rejection[REJECTION_GRAYLISTED];
            log_write_rejection(current_settings, current_settings->transient_rejection, outbound_fd, NULL);

            return_value = FILTER_FLAG_INTERCEPT;
            }
          else if (continue_processing)
            syslog(LOG_ERR, SYSLOG_ERROR_GRAYLIST_FILE "%s: %s", current_settings->graylist_path, strerror(errno));
          }
        }
      }

    if (current_settings->use_syslog)
      syslog(LOG_INFO, "%s from: %s to: %s origin: %s %s %s", (current_settings->transient_rejection != NULL) ? current_settings->transient_rejection->short_reject_message : ((current_settings->rejection != NULL) ? current_settings->rejection->short_reject_message : SHORT_SUCCESS), (current_settings->sender_address[0] != '\0') ? current_settings->sender_address : SYSLOG_MISSING_DATA, (current_settings->recipient_address[0] != '\0') ? current_settings->recipient_address : SYSLOG_MISSING_DATA, (current_settings->server_ip != NULL) ? current_settings->server_ip : SYSLOG_MISSING_DATA, (current_settings->strlen_server_name > 0) ? current_settings->server_name : SYSLOG_MISSING_DATA, (current_settings->smtp_auth_username[0] != '\0') ? current_settings->smtp_auth_username : SYSLOG_MISSING_DATA);

    current_settings->transient_rejection = NULL;

    if ((return_value & FILTER_MASK_PASS) == FILTER_FLAG_PASS)
      current_settings->num_rcpt_to++;
    }
  else if ((strlen_input_line >= STRLEN(SMTP_DATA)) &&
           (strncasecmp(SMTP_DATA, input_line, STRLEN(SMTP_DATA)) == 0))
    {
    if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
      current_settings->filter_action = FILTER_DECISION_DO_FILTER;

    if ((current_settings->filter_action != FILTER_DECISION_DO_FILTER) ||
        (current_settings->rejection == NULL))
      current_settings->inside_data = 1;
    else
      {
      log_write_rejection(current_settings, current_settings->rejection, outbound_fd, NULL);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_QUIT)) &&
           (strncasecmp(SMTP_QUIT, input_line, STRLEN(SMTP_QUIT)) == 0))
    {
    if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
      current_settings->filter_action = FILTER_DECISION_DO_FILTER;

    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->rejection != NULL))
      {
      log_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_QUIT);
      return_value = FILTER_FLAG_QUIT;
      }
    }
  else if ((strlen_input_line >= STRLEN(SMTP_RSET)) &&
           (strncasecmp(SMTP_RSET, input_line, STRLEN(SMTP_RSET)) == 0))
    {
    if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
      current_settings->filter_action = FILTER_DECISION_DO_FILTER;

    if ((current_settings->filter_action != FILTER_DECISION_DO_FILTER) ||
        (current_settings->rejection == NULL))
      {
      current_settings->num_rcpt_to = 0;
      current_settings->local_sender = 1;
      current_settings->sender_address[0] = '\0';
      current_settings->recipient_address[0] = '\0';
      current_settings->graylist_path[0] = '\0';
      }
    else
      {
      log_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_PERMANENT);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }
  else if ((strlen_input_line < STRLEN(SMTP_QUIT)) ||
           (strncasecmp(SMTP_QUIT, input_line, STRLEN(SMTP_QUIT)) != 0))
    {
    if (current_settings->filter_action == FILTER_DECISION_CHECK_AUTH)
      current_settings->filter_action = FILTER_DECISION_DO_FILTER;

    if ((current_settings->filter_action == FILTER_DECISION_DO_FILTER) &&
        (current_settings->rejection != NULL))
      {
      log_write_rejection(current_settings, current_settings->rejection, outbound_fd, REJECT_SEVERITY_PERMANENT);
      return_value = FILTER_FLAG_INTERCEPT | FILTER_FLAG_CHILD_QUIT;
      }
    }

  return(return_value);
  }

void middleman(int *inbound_fd, int *outbound_fd, char line_terminator, struct filter_settings *current_settings, int child_pid, char *envp[])
  {
  char tmp_char;
  char buf_input[MAX_BUF + 1];
  char buf_childinput[MAX_BUF + 1];
  int strlen_buf_input;
  int strlen_buf_childinput;
  int strlen_buf_trim;
  fd_set listen_set;
  struct timeval listen_timeout;
  int more_stdinput;
  int more_childinput;
  int discard_childinput;
  long connection_start;
  long command_start;
  long current_time;
  int filter_return;
  int network_ready;
  int added_ehlo_continuation;

  strlen_buf_input = 0;
  strlen_buf_childinput = 0;
  filter_return = FILTER_FLAG_PASS;

  more_stdinput = 1;
  more_childinput = ((*inbound_fd >= 0) && (*outbound_fd >= 0)) ? 1 : 0;
  discard_childinput = !more_childinput;

  connection_start = time(NULL);
  command_start = time(NULL);

  while ((more_childinput ||
         discard_childinput) &&
         more_stdinput)
    {
    current_time = time(NULL);

    if (((current_settings->timeout_connection > 0) &&
         ((current_time - connection_start) > current_settings->timeout_connection)) ||
        ((current_settings->timeout_command > 0) &&
         ((current_time - command_start) > current_settings->timeout_command)))
      {
      if (*outbound_fd >= 0)
        {
        log_writeln(current_settings, LOG_ACTION_FILTER_TO, *outbound_fd, CHILD_QUIT, STRLEN(CHILD_QUIT));
        close(*outbound_fd);
        *outbound_fd = -1;
        }

      log_write_rejection(current_settings, &rejection[REJECTION_TIMEOUT], STDOUT_FD, NULL);

      more_stdinput = 0;
      discard_childinput = 1;

      if (current_settings->use_syslog)
        syslog(LOG_INFO, "%s from: %s to: %s origin: %s %s %s %s", rejection[REJECTION_TIMEOUT].short_reject_message, (current_settings->sender_address[0] != '\0') ? current_settings->sender_address : SYSLOG_MISSING_DATA, (current_settings->recipient_address[0] != '\0') ? current_settings->recipient_address : SYSLOG_MISSING_DATA, current_settings->server_ip, (current_settings->strlen_server_name > 0) ? current_settings->server_name : SYSLOG_MISSING_DATA, (current_settings->smtp_auth_username[0] != '\0') ? current_settings->smtp_auth_username : SYSLOG_MISSING_DATA, (current_settings->rejection != NULL) ? current_settings->rejection->short_reject_message : SYSLOG_MISSING_DATA);
      }

    listen_timeout.tv_sec = 0;
    listen_timeout.tv_usec = 0;

    if ((current_settings->timeout_connection > 0) ||
        (current_settings->timeout_command > 0))
      {
      if (current_settings->timeout_command > 0)
        listen_timeout.tv_sec = MAXVAL(0, current_settings->timeout_command - (current_time - command_start));

      if (current_settings->timeout_connection > 0)
        listen_timeout.tv_sec = MAXVAL(0, MINVAL(listen_timeout.tv_sec, current_settings->timeout_connection - (current_time - connection_start)));

      if (listen_timeout.tv_sec == 0)
        listen_timeout.tv_usec = 500000;
      }
    else
      listen_timeout.tv_sec = 10;

    FD_ZERO(&listen_set);
    if (more_stdinput)
      FD_SET(STDIN_FD, &listen_set);
    if (more_childinput)
      FD_SET(*inbound_fd, &listen_set);

    if (((network_ready = NETWORK_CAN_READ(current_settings)) > 0) ||
        (select(MAXVAL(STDIN_FD, *inbound_fd) + 1, &listen_set, NULL, NULL, &listen_timeout) > 0))
      {
      if (network_ready)
        FD_ZERO(&listen_set);

      if ((*inbound_fd >= 0) &&
          FD_ISSET(*inbound_fd, &listen_set))
        {
        if (read(*inbound_fd, &tmp_char, 1) > 0)
          {
          if (!discard_childinput)
            {
            if ((line_terminator != '\0') &&
                (((filter_return & FILTER_MASK_AUTH) != FILTER_FLAG_AUTH_NONE) ||
                 ((filter_return & FILTER_MASK_TLS) != FILTER_FLAG_TLS_NONE)))
              {
              buf_childinput[strlen_buf_childinput] = tmp_char;
              strlen_buf_childinput++;

              strlen_buf_trim = strlen_buf_childinput;
              while ((strlen_buf_trim > 0) &&
                     ((buf_childinput[strlen_buf_trim - 1] == '\r') ||
                      (buf_childinput[strlen_buf_trim - 1] == '\n')))
                strlen_buf_trim--;

              if (((strlen_buf_trim > 0) &&
                   (tmp_char == line_terminator)) ||
                  (strlen_buf_trim >= MAX_BUF))
                {
                if ((((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) ||
                     ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD) ||
                     ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_REMOVE)) &&
                    (strlen_buf_trim >= STRLEN(SMTP_EHLO_SUCCESS)) &&
                    (strncasecmp(buf_childinput, SMTP_EHLO_SUCCESS, STRLEN(SMTP_EHLO_SUCCESS)) == 0))
                  {
                  // The server said "250".
                  added_ehlo_continuation = 0;

                  if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) &&
                      (strlen_buf_trim >= (STRLEN(SMTP_EHLO_SUCCESS) + STRLEN(SMTP_EHLO_AUTH) + 1)) &&
                      (strncasecmp(buf_childinput + STRLEN(SMTP_EHLO_SUCCESS) + 1, SMTP_EHLO_AUTH, STRLEN(SMTP_EHLO_AUTH)) == 0))
                    {
                    // The server said "250 AUTH ".  Hopefully it doesn't support an algorithm spamdyke can't handle...
                    if (((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD) &&
                        (buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] != SMTP_CONTINUATION))
                      {
                      buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] = SMTP_CONTINUATION;
                      added_ehlo_continuation = 1;
                      }

                    strlen_buf_trim = 0;
                    current_settings->smtp_auth_origin = SMTP_AUTH_ORIGIN_CHILD;
                    filter_return ^= FILTER_FLAG_AUTH_ADD;
                    log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                    }
                  else if (((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD) &&
                          (strlen_buf_trim >= (STRLEN(SMTP_EHLO_SUCCESS) + STRLEN(SMTP_EHLO_TLS) + 1)) &&
                          (strncasecmp(buf_childinput + STRLEN(SMTP_EHLO_SUCCESS) + 1, SMTP_EHLO_TLS, STRLEN(SMTP_EHLO_TLS)) == 0))
                    {
                    // The server said "250 STARTTLS".
                    if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) &&
                        (buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] != SMTP_CONTINUATION))
                      {
                      buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] = SMTP_CONTINUATION;
                      added_ehlo_continuation = 1;
                      }

                    strlen_buf_trim = 0;
                    filter_return ^= FILTER_FLAG_TLS_ADD;
                    log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                    }
                  else if (((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_REMOVE) &&
                          (strlen_buf_trim >= (STRLEN(SMTP_EHLO_SUCCESS) + STRLEN(SMTP_EHLO_TLS) + 1)) &&
                          (strncasecmp(buf_childinput + STRLEN(SMTP_EHLO_SUCCESS) + 1, SMTP_EHLO_TLS, STRLEN(SMTP_EHLO_TLS)) == 0))
                    {
                    // The server said "250 STARTTLS".  Remove it.
                    log_writeln(current_settings, LOG_ACTION_CHILD_FROM_DISCARDED, -1, buf_childinput, strlen_buf_childinput);

                    if (((filter_return & FILTER_MASK_AUTH) != FILTER_FLAG_AUTH_ADD) &&
                        (buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] != SMTP_CONTINUATION))
                      {
                      // The server said "250 STARTTLS".  Remove it and replace it with something bogus because the previous line was a continuation.
                      log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_SUCCESS, STRLEN(SMTP_EHLO_SUCCESS));
                      log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_STR_DONE, STRLEN(SMTP_STR_DONE));
                      log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_NOTHING_INSERT, STRLEN(SMTP_EHLO_NOTHING_INSERT));
                      }

                    strlen_buf_trim = 0;
                    filter_return ^= FILTER_FLAG_TLS_REMOVE;
                    }

                  if (strlen_buf_trim > 0)
                    {
                    // This line is not "AUTH" or "STARTTLS".
                    if ((strlen_buf_trim >= (STRLEN(SMTP_EHLO_SUCCESS) + 1)) &&
                        (buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] != SMTP_CONTINUATION) &&
                        (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) ||
                         ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD)))
                      {
                      // The server said "250 " without saying "AUTH" and/or "STARTTLS", so change the last line to "250-".
                      buf_childinput[STRLEN(SMTP_EHLO_SUCCESS)] = SMTP_CONTINUATION;
                      added_ehlo_continuation = 1;
                      }

                    log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                    strlen_buf_trim = 0;
                    }

                  if ((((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD) ||
                       ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD) ||
                       ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_REMOVE)) &&
                       added_ehlo_continuation)
                    {
                    if ((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_ADD)
                      {
                      filter_return ^= FILTER_FLAG_AUTH_ADD;
                      current_settings->smtp_auth_origin = SMTP_AUTH_ORIGIN_SPAMDYKE;

                      // Add a "250 AUTH" (or "250-AUTH") line.
                      log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_SUCCESS, STRLEN(SMTP_EHLO_SUCCESS));

                      if ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD)
                        log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_STR_CONTINUATION, STRLEN(SMTP_STR_CONTINUATION));
                      else
                        log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_STR_DONE, STRLEN(SMTP_STR_DONE));

                      if (current_settings->smtp_auth_encryption)
                        log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_AUTH_INSERT_ENCRYPTION, STRLEN(SMTP_EHLO_AUTH_INSERT_ENCRYPTION));
                      else
                        log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_AUTH_INSERT_CLEAR, STRLEN(SMTP_EHLO_AUTH_INSERT_CLEAR));
                      }

                    if ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_ADD)
                      {
                      filter_return ^= FILTER_FLAG_TLS_ADD;

                      // Add a "250 STARTTLS" line.
                      log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_SUCCESS, STRLEN(SMTP_EHLO_SUCCESS));
                      log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_STR_DONE, STRLEN(SMTP_STR_DONE));
                      log_writeln(current_settings, LOG_ACTION_FILTER_FROM, STDOUT_FD, SMTP_EHLO_TLS_INSERT, STRLEN(SMTP_EHLO_TLS_INSERT));
                      }
                    else if ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_REMOVE)
                      filter_return ^= FILTER_FLAG_TLS_REMOVE;
                    }
                  else if (strlen_buf_trim > 0)
                    log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                  }
                else if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE) &&
                         (current_settings->smtp_auth_state == SMTP_AUTH_STATE_CMD_SEEN) &&
                         (strlen_buf_trim >= STRLEN(SMTP_AUTH_CHALLENGE)) &&
                         (strncasecmp(buf_childinput, SMTP_AUTH_CHALLENGE, STRLEN(SMTP_AUTH_CHALLENGE)) == 0))
                  {
                  strncpy(current_settings->smtp_auth_challenge, buf_childinput, MINVAL(MAX_BUF, strlen_buf_trim));
                  current_settings->smtp_auth_challenge[MINVAL(MAX_BUF, strlen_buf_trim)] = '\0';
                  current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_1_SENT;
                  log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                  }
                else if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE) &&
                         (current_settings->smtp_auth_state == SMTP_AUTH_STATE_RESPONSE_1_SEEN) &&
                         (strlen_buf_trim >= STRLEN(SMTP_AUTH_CHALLENGE)) &&
                         (strncasecmp(buf_childinput, SMTP_AUTH_CHALLENGE, STRLEN(SMTP_AUTH_CHALLENGE)) == 0))
                  {
                  current_settings->smtp_auth_state = SMTP_AUTH_STATE_CHALLENGE_2_SENT;
                  log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                  }
                else if (((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE) &&
                         (current_settings->smtp_auth_state == SMTP_AUTH_STATE_RESPONSE_2_SEEN) &&
                         (strlen_buf_trim >= STRLEN(SMTP_AUTH_SUCCESS)) &&
                         (strncasecmp(buf_childinput, SMTP_AUTH_SUCCESS, STRLEN(SMTP_AUTH_SUCCESS)) == 0))
                  {
                  current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
                  current_settings->filter_action = FILTER_DECISION_BYPASS_FILTER;
                  current_settings->rejection = NULL;
                  filter_return ^= FILTER_FLAG_AUTH_CAPTURE;
                  log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                  }
                else if (((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_CAPTURE) &&
                         (strlen_buf_trim >= STRLEN(SMTP_TLS_SUCCESS)) &&
                         (strncasecmp(buf_childinput, SMTP_TLS_SUCCESS, STRLEN(SMTP_TLS_SUCCESS)) == 0))
                  {
                  log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                  current_settings->tls_state = TLS_STATE_ACTIVE_PASSTHROUGH;
                  current_settings->filter_action = FILTER_DECISION_PASS_TLS;
                  line_terminator = '\0';
                  filter_return ^= FILTER_FLAG_TLS_CAPTURE;
                  log_writeln(current_settings, LOG_ACTION_TLS_PASSTHROUGH_START, -1, NULL, -1);
                  }
                else
                  {
                  if ((filter_return & FILTER_MASK_AUTH) == FILTER_FLAG_AUTH_CAPTURE)
                    {
                    filter_return ^= FILTER_FLAG_AUTH_CAPTURE;
                    current_settings->smtp_auth_state = SMTP_AUTH_STATE_NONE;
                    current_settings->filter_action = FILTER_DECISION_DO_FILTER;
                    }
                  else if ((filter_return & FILTER_MASK_TLS) == FILTER_FLAG_TLS_CAPTURE)
                    {
                    filter_return ^= FILTER_FLAG_TLS_CAPTURE;
                    current_settings->filter_action = FILTER_DECISION_DO_FILTER;
                    }

                  log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                  }

                strlen_buf_childinput = 0;
                }
              }
            else
              {
              if (strlen_buf_childinput > 0)
                {
                log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, buf_childinput, strlen_buf_childinput);
                strlen_buf_childinput = 0;
                }

              log_writeln(current_settings, LOG_ACTION_CHILD_FROM, STDOUT_FD, &tmp_char, 1);
              }
            }
          else
            {
            log_writeln(current_settings, LOG_ACTION_CHILD_FROM_DISCARDED, -1, &tmp_char, 1);
            }
          }
        else
          {
          more_childinput = 0;
          close(*inbound_fd);
          *inbound_fd = -1;
          }
        }

      if (FD_ISSET(STDIN_FD, &listen_set) ||
          NETWORK_CAN_READ(current_settings))
        {
        if (NETWORK_READ(current_settings, STDIN_FD, &tmp_char, 1) > 0)
          {
          command_start = time(NULL);

          if (line_terminator != '\0')
            {
            buf_input[strlen_buf_input] = tmp_char;
            strlen_buf_input++;

            if ((tmp_char == line_terminator) ||
                (strlen_buf_input >= MAX_BUF))
              {
              log_writeln(current_settings, LOG_ACTION_REMOTE_FROM, -1, buf_input, strlen_buf_input);

              strlen_buf_trim = strlen_buf_input;
              while ((strlen_buf_trim > 0) &&
                     ((buf_input[strlen_buf_trim - 1] == '\r') ||
                      (buf_input[strlen_buf_trim - 1] == '\n')))
                strlen_buf_trim--;

              filter_return = (tmp_char == line_terminator) ? smtp_filter(STDIN_FD, STDOUT_FD, buf_input, strlen_buf_trim, current_settings, envp) : FILTER_FLAG_PASS;
              switch (filter_return & FILTER_MASK_PASS)
                {
                case FILTER_FLAG_PASS:
                  log_writeln(current_settings, LOG_ACTION_NONE, *outbound_fd, buf_input, strlen_buf_input);
                  break;
                case FILTER_FLAG_QUIT:
                  more_stdinput = 0;
                  discard_childinput = 1;

                  break;
                case FILTER_FLAG_INTERCEPT:
                  if (((filter_return & FILTER_MASK_CHILD) == FILTER_FLAG_CHILD_QUIT) &&
                      (current_settings->rejection != NULL) &&
                      (*outbound_fd >= 0) &&
                      !discard_childinput)
                    {
                    log_writeln(current_settings, LOG_ACTION_FILTER_TO, *outbound_fd, CHILD_QUIT, STRLEN(CHILD_QUIT));
                    close(*outbound_fd);
                    *outbound_fd = -1;

                    discard_childinput = 1;
                    }
                }

              strlen_buf_input = 0;
              }
            }
          else
            log_writeln(current_settings, LOG_ACTION_REMOTE_FROM, *outbound_fd, &tmp_char, 1);
          }
        else
          {
          if (*outbound_fd >= 0)
            {
            if (strlen_buf_input > 0)
              {
              log_writeln(current_settings, LOG_ACTION_REMOTE_FROM, *outbound_fd, buf_input, strlen_buf_input);
              strlen_buf_input = 0;
              }

            close(*outbound_fd);
            *outbound_fd = -1;
            }

          more_stdinput = 0;
          }
        }
      }

    if (child_pid != -1)
      waitpid(child_pid, NULL, WNOHANG);
    }

  return;
  }

int search_ip(char *target_ip, int strlen_target_ip, char *target_name, int strlen_target_name)
  {
  int j;
  int k;

  j = 0;
  k = 0;
  while (k < strlen_target_name)
    if (j < strlen_target_ip)
      if ((target_ip[j] >= '0') &&
          (target_ip[j] <= '9'))
        {
        if (target_ip[j] == target_name[k])
          j++;
        else
          j = 0;

        k++;
        }
      else
        {
        while ((k < strlen_target_name) &&
               ((target_name[k] < '0') ||
                (target_name[k] > '9')))
          k++;

        j++;
        }
    else
      break;

  return(((j > 0) && (j == strlen_target_ip)) ? 1 : 0);
  }

int is_ip_in_name(char *target_ip, char *target_name)
  {
  int return_value;
  char ip_octets[4][4];
  int ip_ints[4];
  char tmp_ip[MAX_IP + 1];
  int tmp_strlen;
  struct in_addr tmp_addr;
  int strlen_target_ip;
  int strlen_target_name;

  return_value = -1;

  if ((target_ip != NULL) &&
      ((strlen_target_ip = strlen(target_ip)) > 0) &&
      (sscanf(target_ip, "%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]) == 4) &&
      (sscanf(ip_octets[0], "%d", &ip_ints[0]) == 1) &&
      (ip_ints[0] >= 0) &&
      (ip_ints[0] <= 255) &&
      (sscanf(ip_octets[1], "%d", &ip_ints[1]) == 1) &&
      (ip_ints[1] >= 0) &&
      (ip_ints[1] <= 255) &&
      (sscanf(ip_octets[2], "%d", &ip_ints[2]) == 1) &&
      (ip_ints[2] >= 0) &&
      (ip_ints[2] <= 255) &&
      (sscanf(ip_octets[3], "%d", &ip_ints[3]) == 1) &&
      (ip_ints[3] >= 0) &&
      (ip_ints[3] <= 255) &&
      (target_name != NULL) &&
      ((strlen_target_name = strlen(target_name)) > 0))
    {
    return_value = 0;

//printf("Looking for %s\n", target_ip); //FIXME
    // This block looks for the dotted-quad IP address in the rDNS name.
    return_value = search_ip(target_ip, strlen_target_ip, target_name, strlen_target_name);

    // This block looks for the zero-padded dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 011.022.033.044
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%.3d.%.3d.%.3d.%.3d", ip_ints[0], ip_ints[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the reversed dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 44.33.22.11.
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%s.%s", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the slightly out-of-order dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 44.11.22.33
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%s.%s", ip_octets[3], ip_octets[0], ip_octets[1], ip_octets[2]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the strangely out-of-order dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 33.22.11.44
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%s.%s", ip_octets[2], ip_octets[1], ip_octets[0], ip_octets[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the strangely out-of-order dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 44.33.1122
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%s%s", ip_octets[3], ip_octets[2], ip_octets[0], ip_octets[1]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the strangely out-of-order dotted-quad IP address in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 3344.11.22
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s%s.%s.%s", ip_octets[2], ip_octets[3], ip_octets[0], ip_octets[1]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the dotted-quad IP address with the last two octets encoded as a long integer in the rDNS
    // name.  For example, if the IP is 11.22.33.44, this block looks for
    // 11.22.8492
    if (!return_value)
      {
      tmp_strlen = snprintf(tmp_ip, MAX_IP, "%s.%s.%d", ip_octets[0], ip_octets[1], (ip_ints[2] * 256) + ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME
      return_value = search_ip(tmp_ip, tmp_strlen, target_name, strlen_target_name);
      }

    // This block looks for the undotted-quad IP address in the rDNS name.
    // For example, if the IP is 11.22.33.44, this block looks for
    // 11223344.
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%s%s%s", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, zero-padded-quad IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 011022033044
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%.3d%.3d%.3d%.3d", ip_ints[0], ip_ints[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, partially zero-padded-quad IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 11022033044
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%.3d%.3d%.3d", ip_octets[0], ip_ints[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, partially zero-padded-quad IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 1122033044
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%s%.3d%.3d", ip_octets[0], ip_octets[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, partially zero-padded-quad IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 112233044
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%s%s%.3d", ip_octets[0], ip_octets[1], ip_octets[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-quad reversed IP address in the rDNS name.
    // For example, if the IP is 11.22.33.44, this block looks for
    // 44332211.
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%s%s%s%s", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the undotted-, zero-padded-quad reversed IP address in the
    // rDNS name.  For example, if the IP is 11.22.33.44, this block looks for
    // 044033022011
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%.3d%.3d%.3d%.3d", ip_ints[3], ip_ints[2], ip_ints[1], ip_ints[0]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the IP address as an unsigned long integer.
    // For example, if the IP is 85.135.72.234, this block looks for
    // 3930621781.
    if (!return_value &&
        inet_aton(target_ip, &tmp_addr))
      {
      snprintf(tmp_ip, MAX_IP, "%lu", (long unsigned int)tmp_addr.s_addr);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }

    // This block looks for the IP address as a sequence of hex bytes.
    // For example, if the IP is 85.135.72.234, this block looks for
    // 5080d7e3.
    if (!return_value)
      {
      snprintf(tmp_ip, MAX_IP, "%.2x%.2x%.2x%.2x", ip_ints[0], ip_ints[1], ip_ints[2], ip_ints[3]);
//printf("Looking for %s\n", tmp_ip); //FIXME

      if (strstr(target_name, tmp_ip) != NULL)
        return_value = 1;
      }
    }

  return(return_value);
  }

int check_rdns_resolve(char *target_hostname)
  {
  int return_value;

  return_value = 0;

  if ((target_hostname != NULL) &&
      (target_hostname[0] != '\0'))
    {
    if ((gethostbyname(target_hostname) != NULL) ||
        ((h_errno != HOST_NOT_FOUND) &&
         (h_errno != NO_ADDRESS) &&
         (h_errno != NO_DATA)))
      return_value = 1;
    }
  else
    return_value = 1;

  return(return_value);
  }

int check_country_code(char *target_name, int strlen_target_name)
  {
  int return_value;
  int j;

  return_value = 0;

  if ((target_name != NULL) &&
      (strlen_target_name > 0))
    for (j = (strlen_target_name - 1); j >= 0; j--)
      if (target_name[j] == '.')
        {
        if ((strlen_target_name - j) == 3)
          return_value = 1;

        break;
        }

  return(return_value);
  }

int check_rdns_keywords(char *keywords_filename, char *target_name, int strlen_target_name)
  {
  int return_value;
  FILE *tmp_file;
  char tmp_buf[MAX_FILE_BUF + 1];
  char tmp_pattern[MAX_BUF + 1];
  int i;
  int j;
  char *tld_start;
  int strlen_keyword;
  int strlen_subkeyword;
  char *tmp_char;
  int found;
  int not_found;

  return_value = 0;

  if ((target_name != NULL) &&
      (strlen_target_name > 0))
    {
    if ((tmp_file = fopen(keywords_filename, "r")) != NULL)
      {
      // It's important to know where the domain name begins, since
      // disqualifying mail.dhcp.com isn't fair.
      tld_start = NULL;
      for (j = (strlen_target_name - 1); j >= 0; j--)
        if (target_name[j] == '.')
          {
          for (j--; j >= 0; j--)
            if (target_name[j] == '.')
              {
              tld_start = target_name + j + 1;
              break;
              }

          break;
          }

      if (tld_start != NULL)
        {
        i = 0;

        while (!feof(tmp_file) &&
               (i < MAX_FILE_LINES))
          {
          if ((snprintf(tmp_pattern, MAX_BUF, "%%%d[^\r\n]%%*[\r\n]", MAX_FILE_BUF) > 0) &&
              (fscanf(tmp_file, tmp_pattern, tmp_buf) == 1) &&
              (tmp_buf[0] != '\0') &&
              (tmp_buf[0] != '#') &&
              ((strlen_keyword = strlen(tmp_buf)) > 0))
            {
            for (j = 0; j < strlen_keyword; j++)
              if (tmp_buf[j] == ' ')
                tmp_buf[j] = '\0';

            j = 0;
            not_found = 0;
            while (j < strlen_keyword)
              {
              strlen_subkeyword = strlen(tmp_buf + j);

              found = 0;
              tmp_char = target_name;
              while (((tmp_char - target_name) < strlen_target_name) &&
                     ((tmp_char = strstr(tmp_char, tmp_buf + j)) != NULL) &&
                     (tmp_char < tld_start))
                {
                // The preceding and following characters around the match must be
                // non-alphanumeric so we don't reject 10.0.0.1.notdialup.example.com.
                // This means 10.0.0.1.dialup3.example.com will get through but the
                // alternative is worse.
                if (((tmp_char == target_name) ||
                     !isalnum((tmp_buf + j)[0]) ||
                     !isalnum((tmp_char - 1)[0])) &&
                    (!isalnum((tmp_buf + j + strlen_subkeyword - 1)[0]) ||
                     !isalnum((tmp_char + strlen_subkeyword)[0])))
                  {
                  found = 1;
                  break;
                  }

                tmp_char += strlen_subkeyword;
                }

              if (!found)
                {
                not_found = 1;
                break;
                }

              while (j < strlen_keyword)
                if (tmp_buf[j++] == '\0')
                  break;
              }

            if (!not_found)
              {
              return_value = 1;
              break;
              }
            }

          i++;
          }

        if (i == MAX_FILE_LINES)
          syslog(LOG_ERR, SYSLOG_ERROR_FILE_TOO_LONG "%s", MAX_FILE_LINES, keywords_filename);

        fclose(tmp_file);
        }
      }
    else
      syslog(LOG_ERR, SYSLOG_ERROR_OPEN_KEYWORDS "%s: %s", keywords_filename, strerror(errno));
    }

  return(return_value);
  }

int check_earlytalker(int delay)
  {
  int return_value;
  struct timeval tmp_time;
  fd_set listen_set;

  return_value = 0;

  tmp_time.tv_sec = delay;
  tmp_time.tv_usec = 0;

  FD_ZERO(&listen_set);
  FD_SET(STDIN_FD, &listen_set);
  if (select(STDIN_FD + 1, &listen_set, NULL, NULL, &tmp_time) == 1)
    return_value = 1;

  return(return_value);
  }

int run_tests(struct filter_settings *current_settings, char *envp[])
  {
  static char *environment_hostname[] = ENVIRONMENT_HOSTNAME;
  static int strlen_environment_hostname[] = STRLEN_ENVIRONMENT_HOSTNAME;
  int return_value;
  int i;
  int ip_in_name;
  char rbl_name[MAX_BUF + 1];
  char ip_octets[4][4];

  return_value = FILTER_DECISION_UNDECIDED;
  ip_in_name = -1;

  // Check for rDNS name
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      current_settings->check_rdns_exist &&
      (current_settings->strlen_server_name == 0))
    {
    current_settings->rejection = &rejection[REJECTION_RDNS_MISSING];
    return_value = FILTER_DECISION_DO_FILTER;
    }

  // Check for IP and CC in name
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      current_settings->check_ip_in_rdns_cc &&
      (current_settings->strlen_server_name > 0))
    {
    if (ip_in_name == -1)
      ip_in_name = is_ip_in_name(current_settings->server_ip, current_settings->server_name);

    if (ip_in_name &&
        check_country_code(current_settings->server_name, current_settings->strlen_server_name))
      {
      current_settings->rejection = &rejection[REJECTION_IP_IN_NAME_CC];
      return_value = FILTER_DECISION_DO_FILTER;
      }
    }

  // Check rDNS blacklist
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->blacklist_rdns_directory_path != NULL) &&
      (current_settings->strlen_server_name > 0))
    for (i = 0; current_settings->blacklist_rdns_directory_path[i] != NULL; i++)
      if (search_domain_directory(current_settings->blacklist_rdns_directory_path[i], current_settings->server_name, current_settings->strlen_server_name) == 1)
        {
        current_settings->rejection = &rejection[REJECTION_BLACKLIST_NAME];
        return_value = FILTER_DECISION_DO_FILTER;
        break;
        }

  // Check rDNS whitelist
  if ((current_settings->whitelist_rdns_path != NULL) &&
      (current_settings->strlen_server_name > 0))
    for (i = 0; current_settings->whitelist_rdns_path[i] != NULL; i++)
      if (search_file(current_settings->whitelist_rdns_path[i], current_settings->server_name, current_settings->strlen_server_name, '.', '\0') == 1)
        {
        current_settings->rejection = NULL;
        return_value = FILTER_DECISION_DO_NOT_FILTER;
        break;
        }

  // Check IP whitelist
  if (current_settings->whitelist_ip_path != NULL)
    for (i = 0; current_settings->whitelist_ip_path[i] != NULL; i++)
      if (search_tcprules_file(NULL, 0, current_settings->whitelist_ip_path[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name) == 1)
        {
        current_settings->rejection = NULL;
        return_value = FILTER_DECISION_DO_NOT_FILTER;
        break;
        }

  // Check IP blacklist
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->blacklist_ip_path != NULL))
    for (i = 0; current_settings->blacklist_ip_path[i] != NULL; i++)
      if (search_tcprules_file(NULL, 0, current_settings->blacklist_ip_path[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name) == 1)
        {
        current_settings->rejection = &rejection[REJECTION_BLACKLIST_IP];
        return_value = FILTER_DECISION_DO_FILTER;
        break;
        }

  // Check for IP and keywords in name
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->blacklist_rdns_keyword_path != NULL) &&
      (current_settings->strlen_server_name > 0))
    {
    if (ip_in_name == -1)
      ip_in_name = is_ip_in_name(current_settings->server_ip, current_settings->server_name);

    if (ip_in_name)
      for (i = 0; current_settings->blacklist_rdns_keyword_path[i] != NULL; i++)
        if (check_rdns_keywords(current_settings->blacklist_rdns_keyword_path[i], current_settings->server_name, current_settings->strlen_server_name))
          {
          current_settings->rejection = &rejection[REJECTION_IP_IN_NAME];
          return_value = FILTER_DECISION_DO_FILTER;
          break;
          }
    }

  // Check for rDNS resolution
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      current_settings->check_rdns_resolve &&
      (current_settings->strlen_server_name > 0) &&
      (((strncasecmp(current_settings->server_name, LOCALHOST_NAME, STRLEN(LOCALHOST_NAME)) != 0) &&
        !check_rdns_resolve(current_settings->server_name)) ||
       ((strncasecmp(current_settings->server_name, LOCALHOST_NAME, STRLEN(LOCALHOST_NAME)) == 0) &&
        (strncasecmp(current_settings->server_ip, LOCALHOST_IP, STRLEN(LOCALHOST_IP)) != 0))))
    {
    current_settings->rejection = &rejection[REJECTION_RDNS_RESOLVE];
    return_value = FILTER_DECISION_DO_FILTER;
    }

  // Check DNS RBL
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->dnsrbl_fqdn != NULL) &&
      (strcmp(current_settings->server_ip, LOCALHOST_IP) != 0) &&
      (sscanf(current_settings->server_ip, "%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]) == 4))
    for (i = 0; current_settings->dnsrbl_fqdn[i] != NULL; i++)
      if (snprintf(rbl_name, MAX_BUF, "%s.%s.%s.%s.%s", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0], current_settings->dnsrbl_fqdn[i]))
        {
        if (dns_txt(rbl_name, current_settings))
          {
          return_value = FILTER_DECISION_DO_FILTER;
          break;
          }
        else if (gethostbyname(rbl_name) != NULL)
          {
          memcpy(&current_settings->rejection_buf, &rejection[REJECTION_RBL], sizeof(struct rejection_data));
          current_settings->rejection_buf.strlen_reject_message = snprintf(current_settings->reject_message_buf, MAX_BUF, rejection[REJECTION_RBL].reject_message, current_settings->dnsrbl_fqdn[i]);
          current_settings->rejection_buf.reject_message = current_settings->reject_message_buf;
          current_settings->rejection = &current_settings->rejection_buf;

          return_value = FILTER_DECISION_DO_FILTER;
          break;
          }
        }

  // Check for earlytalkers
  if ((return_value == FILTER_DECISION_UNDECIDED) &&
      (current_settings->check_earlytalker > 0) &&
      check_earlytalker(current_settings->check_earlytalker))
    {
    current_settings->rejection = &rejection[REJECTION_EARLYTALKER];
    return_value = FILTER_DECISION_DO_FILTER;
    }

  // Attempt to find the local hostname if possible.
  if (current_settings->local_server_name == NULL)
    for (i = 0; environment_hostname[i] != NULL; i++)
      if ((current_settings->local_server_name = find_environment_variable(envp, environment_hostname[i], strlen_environment_hostname[i])) != NULL)
        break;

  if (((return_value == FILTER_DECISION_DO_FILTER) ||
       (return_value == FILTER_DECISION_UNDECIDED)) &&
      (current_settings->smtp_auth_command != NULL))
    return_value = FILTER_DECISION_CHECK_AUTH;
  else if (return_value == FILTER_DECISION_UNDECIDED)
    return_value = current_settings->filter_action;

  return(return_value);
  }

int process_access(struct filter_settings *current_settings, char *envp[], char ***destination_envp)
  {
  int return_value;
  int i;
  int j;
  int k;
  char command_text[MAX_BUF + 1];
  char *current_ptr;
  char *value_start_ptr;
  char *value_end_ptr;
  int strlen_envp;
  int len_envp;
  char tmp_delimiter;
  void *tmp_ptr;

  return_value = FILTER_DECISION_UNDECIDED;

  if (current_settings->access_list_path != NULL)
    for (i = 0; current_settings->access_list_path[i] != NULL; i++)
      if (search_tcprules_file(command_text, MAX_BUF, current_settings->access_list_path[i], envp, current_settings->server_ip, current_settings->server_name, current_settings->strlen_server_name))
        {
        current_settings->allow_relay = 0;

        if ((strncasecmp(command_text, TCPRULES_ALLOW, STRLEN(TCPRULES_ALLOW)) == 0) &&
            ((command_text[STRLEN(TCPRULES_ALLOW)] == TCPRULES_DELIMITER) ||
             (command_text[STRLEN(TCPRULES_ALLOW)] == '\0')))
          {
          return_value = FILTER_DECISION_DO_FILTER;

          if ((*destination_envp == NULL) ||
              (*destination_envp == envp))
            {
            if (envp != NULL)
              {
              for (len_envp = 0; envp[len_envp] != NULL; len_envp++);
              if ((*destination_envp = malloc(sizeof(char *) * (len_envp + 1))) != NULL)
                {
                (*destination_envp)[0] = NULL;
                for (j = 0; j < len_envp; j++)
                  {
                  strlen_envp = strlen(envp[j]);

                  if (strncasecmp(envp[j], ENVIRONMENT_ALLOW_RELAY, MAXVAL(STRLEN(ENVIRONMENT_ALLOW_RELAY), strlen_envp)) == 0)
                    current_settings->allow_relay = 1;

                  if (((*destination_envp)[j] = malloc(sizeof(char) * (strlen_envp + 1))) != NULL)
                    {
                    strncpy((*destination_envp)[j], envp[j], strlen_envp);
                    (*destination_envp)[j + 1] = NULL;
                    }
                  else
                    {
                    for (k = 0; k < j; k++)
                      free((*destination_envp)[k]);
                    free(*destination_envp);
                    *destination_envp = NULL;

                    syslog(LOG_ERR, SYSLOG_ERROR_REALLOC, (unsigned long)(sizeof(char) * (strlen_envp + 1)));
                    return_value = FILTER_DECISION_ERROR;
                    break;
                    }
                  }
                }
              else
                {
                syslog(LOG_ERR, SYSLOG_ERROR_REALLOC, (unsigned long)(sizeof(char *) * (len_envp + 1)));
                return_value = FILTER_DECISION_ERROR;
                break;
                }
              }
            else
              {
              *destination_envp = NULL;
              len_envp = 0;
              }
            }
          else
            for (len_envp = 0; (*destination_envp)[len_envp] != NULL; len_envp++);

        if (command_text[STRLEN(TCPRULES_ALLOW)] == TCPRULES_DELIMITER)
          {
          current_ptr = command_text + STRLEN(TCPRULES_ALLOW) + 1;
          while (((current_ptr - command_text) < MAX_BUF) &&
                 (current_ptr[0] != '\0'))
            {
            for (value_start_ptr = current_ptr; (((value_start_ptr - command_text) < (MAX_BUF - 3)) && (value_start_ptr[0] != '\0') && (value_start_ptr[0] != ENVIRONMENT_DELIMITER)); value_start_ptr++);
            if (value_start_ptr[0] == ENVIRONMENT_DELIMITER)
              {
              for (j = 0; j < len_envp; j++)
                if (strncasecmp((*destination_envp)[j], current_ptr, value_start_ptr - current_ptr) == 0)
                  break;

              tmp_delimiter = value_start_ptr[1];
              value_start_ptr += 2;
              for (value_end_ptr = value_start_ptr; (((value_end_ptr - command_text) < (MAX_BUF - 1)) && (value_end_ptr[0] != '\0') && (value_end_ptr[0] != tmp_delimiter)); value_end_ptr++);
              if (j == len_envp)
                {
                if ((tmp_ptr = realloc(*destination_envp, sizeof(char *) * (len_envp + 2))) != NULL)
                  {
                  *destination_envp = tmp_ptr;
                  (*destination_envp)[len_envp] = NULL;
                  (*destination_envp)[len_envp + 1] = NULL;
                  len_envp++;
                  }
                else
                  {
                  syslog(LOG_ERR, SYSLOG_ERROR_REALLOC, (unsigned long)(sizeof(char *) * (len_envp + 2)));
                  return_value = FILTER_DECISION_ERROR;
                  break;
                  }
                }

              if ((tmp_ptr = realloc((*destination_envp)[j], sizeof(char) * (value_end_ptr - current_ptr))) != NULL)
                (*destination_envp)[j] = tmp_ptr;
              else
                {
                syslog(LOG_ERR, SYSLOG_ERROR_REALLOC, (unsigned long)(sizeof(char) * ((value_end_ptr - value_start_ptr) + 1)));
                return_value = FILTER_DECISION_ERROR;
                break;
                }

              if (strncasecmp(current_ptr, ENVIRONMENT_ALLOW_RELAY, STRLEN(ENVIRONMENT_ALLOW_RELAY)) == 0)
                current_settings->allow_relay = 1;

              strncpy((*destination_envp)[j], current_ptr, (value_start_ptr - current_ptr) - 1);
              strncpy((*destination_envp)[j] + ((value_start_ptr - current_ptr) - 1), value_start_ptr, value_end_ptr - value_start_ptr);
              (*destination_envp)[j][(value_end_ptr - current_ptr) - 1] = '\0';
              current_ptr = value_end_ptr + 2;
              }
            else
              current_ptr = value_start_ptr;
            }
          }

        if (!current_settings->allow_relay &&
            (current_settings->local_domains_path != NULL))
          {
          if ((tmp_ptr = realloc(*destination_envp, sizeof(char *) * (len_envp + 2))) != NULL)
            {
            *destination_envp = tmp_ptr;
            (*destination_envp)[len_envp] = NULL;
            (*destination_envp)[len_envp + 1] = NULL;
            len_envp++;

            if ((tmp_ptr = realloc((*destination_envp)[len_envp - 1], sizeof(char) * (STRLEN(ENVIRONMENT_ALLOW_RELAY) + 2))) != NULL)
              {
              (*destination_envp)[len_envp - 1] = tmp_ptr;
              strncpy((*destination_envp)[len_envp - 1], ENVIRONMENT_ALLOW_RELAY, STRLEN(ENVIRONMENT_ALLOW_RELAY));
              (*destination_envp)[len_envp - 1][STRLEN(ENVIRONMENT_ALLOW_RELAY)] = ENVIRONMENT_DELIMITER;
              (*destination_envp)[len_envp - 1][STRLEN(ENVIRONMENT_ALLOW_RELAY) + 1] = '\0';
              }
            else
              {
              syslog(LOG_ERR, SYSLOG_ERROR_REALLOC, (unsigned long)(sizeof(char) * (STRLEN(ENVIRONMENT_ALLOW_RELAY) + 2)));
              return_value = FILTER_DECISION_ERROR;
              break;
              }
            }
          else
            {
            syslog(LOG_ERR, SYSLOG_ERROR_REALLOC, (unsigned long)(sizeof(char *) * (len_envp + 2)));
            return_value = FILTER_DECISION_ERROR;
            break;
            }
          }

        break;
        }
      else if ((strncasecmp(command_text, TCPRULES_DENY, STRLEN(TCPRULES_DENY)) == 0) &&
               ((command_text[STRLEN(TCPRULES_DENY)] == TCPRULES_DELIMITER) ||
                (command_text[STRLEN(TCPRULES_DENY)] == '\0')))
        {
        current_settings->rejection = &rejection[REJECTION_ACCESS_DENIED];
        return_value = FILTER_DECISION_DO_FILTER;
        break;
        }
      }

  return(return_value);
  }

int main(int argc, char *argv[], char *envp[])
  {
  struct filter_settings tmp_settings;
  int output_pipe[2];
  int input_pipe[2];
  int child_pid;
  int test_result;
  char **new_envp;

  srandom(time(NULL));
  openlog(SYSLOG_IDENTIFIER, LOG_PID, LOG_MAIL);

  init_settings(&tmp_settings);
  new_envp = envp;

  if ((((tmp_settings.filter_action = process_command_line(&tmp_settings, argc, argv, envp)) == FILTER_DECISION_UNDECIDED) ||
       (tmp_settings.filter_action == FILTER_DECISION_DO_FILTER)) &&
      (((test_result = process_access(&tmp_settings, envp, &new_envp)) == FILTER_DECISION_UNDECIDED) ||
       (test_result == FILTER_DECISION_DO_FILTER)))
    {
    if (tmp_settings.filter_action == FILTER_DECISION_UNDECIDED)
      tmp_settings.filter_action = test_result;

    if ((test_result = run_tests(&tmp_settings, new_envp)) != FILTER_DECISION_UNDECIDED)
      tmp_settings.filter_action = test_result;
    }

#ifdef TLS
  if ((tmp_settings.tls_certificate_path != NULL) &&
      !tls_init(&tmp_settings))
    {
    tmp_settings.tls_certificate_path = NULL;
    tmp_settings.tls_privatekey_path = NULL;
    tmp_settings.tls_privatekey_password = NULL;
    }
#endif /* TLS */

  output_pipe[0] = -1;
  output_pipe[1] = -1;
  input_pipe[0] = -1;
  input_pipe[1] = -1;

  if ((tmp_settings.filter_action == FILTER_DECISION_DO_FILTER) &&
      (tmp_settings.rejection != NULL))
    {
    log_write_rejection(&tmp_settings, tmp_settings.rejection, STDOUT_FD, REJECT_SEVERITY_HELO);

    middleman(&input_pipe[0], &output_pipe[1], '\n', &tmp_settings, -1, new_envp);
    }
  else if ((tmp_settings.filter_action == FILTER_DECISION_DO_FILTER) ||
           (tmp_settings.filter_action == FILTER_DECISION_CHECK_AUTH))
    {
    if (pipe(output_pipe) != -1)
      {
      if (pipe(input_pipe) != -1)
        {
        if ((child_pid = fork()) > 0)
          // parent
          {
          close(output_pipe[0]);
          close(input_pipe[1]);

          middleman(&input_pipe[0], &output_pipe[1], '\n', &tmp_settings, child_pid, new_envp);

          if (output_pipe[1] != -1)
            close(output_pipe[1]);
          if (input_pipe[0] != -1)
            close(input_pipe[0]);
          wait(NULL);
          }
        else if (child_pid == 0)
          // child
          {
          close(output_pipe[1]);
          close(input_pipe[0]);

          if ((dup2(output_pipe[0], STDIN_FD) != -1) &&
              (dup2(input_pipe[1], STDOUT_FD) != -1))
            {
            if (!exec_path(tmp_settings.child_argv[0], tmp_settings.child_argv, new_envp))
              {
              syslog(LOG_ERR, SYSLOG_ERROR_EXEC "%s: %s", tmp_settings.child_argv[0], strerror(errno));
              close(output_pipe[0]);
              close(input_pipe[1]);
              }
            }
          else
            {
            syslog(LOG_CRIT, SYSLOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

            close(output_pipe[0]);
            close(input_pipe[1]);

            tmp_settings.filter_action = FILTER_DECISION_DO_NOT_FILTER;
            }
          }
        else
          {
          syslog(LOG_CRIT, SYSLOG_ERROR_FORK "%s", strerror(errno));

          close(output_pipe[0]);
          close(output_pipe[1]);
          close(input_pipe[0]);
          close(input_pipe[1]);

          tmp_settings.filter_action = FILTER_DECISION_DO_NOT_FILTER;
          }
        }
      else
        {
        syslog(LOG_CRIT, SYSLOG_ERROR_PIPE "%s", strerror(errno));

        close(output_pipe[0]);
        close(output_pipe[1]);

        tmp_settings.filter_action = FILTER_DECISION_DO_NOT_FILTER;
        }
      }
    else
      {
      syslog(LOG_CRIT, SYSLOG_ERROR_PIPE "%s", strerror(errno));
      tmp_settings.filter_action = FILTER_DECISION_DO_NOT_FILTER;
      }
    }

  if (((tmp_settings.filter_action == FILTER_DECISION_UNDECIDED) ||
       (tmp_settings.filter_action == FILTER_DECISION_DO_NOT_FILTER)) &&
      !exec_path(tmp_settings.child_argv[0], tmp_settings.child_argv, new_envp))
    syslog(LOG_ERR, SYSLOG_ERROR_EXEC "%s: %s", tmp_settings.child_argv[0], strerror(errno));

#ifdef TLS
  if (tls_end(&tmp_settings, STDIN_FD))
    log_writeln(&tmp_settings, LOG_ACTION_TLS_END, -1, NULL, -1);
#endif // TLS

  log_writeln(NULL, 0, -1, NULL, 0);
  closelog();

  return(0);
  }
