/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifdef TLS

#include <openssl/rand.h>
#include <openssl/err.h>
#include <errno.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include "spamdyke.h"
#include "tls.h"

int tls_password_callback(char *buf, int size, int rwflag, void *userdata)
  {
  return((((struct filter_settings *)userdata)->tls_privatekey_password != NULL) ? snprintf(buf, size, "%s", ((struct filter_settings *)userdata)->tls_privatekey_password) : 0);
  }

void tls_error(struct filter_settings *current_settings, int return_code, char *syslog_text, char *additional_text)
  {
  char error_text[MAX_BUF + 1];
  int strlen_error_text;
  int saved_errno;
  int tls_error;
  int ssl_error;

  saved_errno = errno;
  strlen_error_text = 0;

  ssl_error = SSL_get_error(current_settings->tls_session, return_code);

  error_text[0] = '\0';
  while (((tls_error = ERR_get_error()) != 0) &&
         (strlen_error_text < MAX_BUF))
    {
    if (strlen_error_text > 0)
      strlen_error_text += snprintf(error_text + strlen_error_text, MAX_BUF - strlen_error_text, ", ");

    ERR_error_string_n(tls_error, error_text + strlen_error_text, MAX_BUF - strlen_error_text);
    strlen_error_text += strlen(error_text + strlen_error_text);
    }

  switch (ssl_error)
    {
    case SSL_ERROR_NONE:
      // No error occurred.
      if (additional_text != NULL)
        syslog(LOG_ERR, "%s (%s): %s", syslog_text, additional_text, SYSLOG_ERROR_TLS_NO_ERROR);
      else
        syslog(LOG_ERR, "%s: %s", syslog_text, SYSLOG_ERROR_TLS_NO_ERROR);

      break;
    case SSL_ERROR_ZERO_RETURN:
      // SSL connection closed
      if (additional_text != NULL)
        syslog(LOG_ERR, "%s (%s): %s%s", syslog_text, additional_text, SYSLOG_ERROR_TLS_ZERO_RETURN, error_text);
      else
        syslog(LOG_ERR, "%s: %s%s", syslog_text, SYSLOG_ERROR_TLS_ZERO_RETURN, error_text);

      break;
    case SSL_ERROR_WANT_READ:
    case SSL_ERROR_WANT_WRITE:
      // operation did not complete, call it again
    case SSL_ERROR_WANT_CONNECT:
    case SSL_ERROR_WANT_ACCEPT:
      // operation did not complete, call it again
    case SSL_ERROR_WANT_X509_LOOKUP:
      // callback function wants another callback.  Call the SSL function again.
      if (additional_text != NULL)
        syslog(LOG_ERR, "%s (%s): %s%s", syslog_text, additional_text, SYSLOG_ERROR_TLS_RECALL, error_text);
      else
        syslog(LOG_ERR, "%s: %s%s", syslog_text, SYSLOG_ERROR_TLS_RECALL, error_text);

      break;
    case SSL_ERROR_SYSCALL:
      // check the SSL error queue.  If return_code == 0, EOF found.  If return_code == -1, check errno.
      if (return_code == 0)
        strlen_error_text += snprintf(error_text + strlen_error_text, MAX_BUF - strlen_error_text, ", %s", SYSLOG_ERROR_TLS_EOF_FOUND);
      else if (return_code == -1)
        strlen_error_text += snprintf(error_text + strlen_error_text, MAX_BUF - strlen_error_text, ", %s", strerror(saved_errno));

      if (additional_text != NULL)
        syslog(LOG_ERR, "%s (%s): %s%s", syslog_text, additional_text, SYSLOG_ERROR_TLS_SYSCALL, error_text);
      else
        syslog(LOG_ERR, "%s: %s%s", syslog_text, SYSLOG_ERROR_TLS_SYSCALL, error_text);

      break;
    case SSL_ERROR_SSL:
      // Library failure, check the SSL error queue.
      if (additional_text != NULL)
        syslog(LOG_ERR, "%s (%s): %s%s", syslog_text, additional_text, SYSLOG_ERROR_TLS_LIBRARY, error_text);
      else
        syslog(LOG_ERR, "%s: %s%s", syslog_text, SYSLOG_ERROR_TLS_LIBRARY, error_text);

      break;
    default:
      if (additional_text != NULL)
        syslog(LOG_ERR, "%s (%s): %s%s", syslog_text, additional_text, SYSLOG_MISSING_DATA, error_text);
      else
        syslog(LOG_ERR, "%s: %s%s", syslog_text, SYSLOG_MISSING_DATA, error_text);

      break;
    }

  return;
  }

int tls_init(struct filter_settings *current_settings)
  {
  static int initialized = 0;
  int return_value;
  int i;
  int tmp_rand;
  int error_occurred;
  int tls_return;

  return_value = 0;

  if (!initialized)
    {
    SSL_library_init();

    if (!RAND_status())
      {
      srand(time(NULL));

      i = 0;
      do
        {
        tmp_rand = rand();
        RAND_seed(&tmp_rand, sizeof(int));
        i++;
        }
      while (!RAND_status() &&
             (i < MAX_RAND_SEED));

      if (i < MAX_RAND_SEED)
        initialized = 1;
      else
        syslog(LOG_ERR, SYSLOG_ERROR_TLS_INIT);
      }
    else
      initialized = 1;
    }

  if (initialized)
    {
    if ((current_settings->tls_context = SSL_CTX_new(SSLv23_server_method())) != NULL)
      {
      error_occurred = 0;

      if (current_settings->tls_privatekey_password != NULL)
        {
        SSL_CTX_set_default_passwd_cb(current_settings->tls_context, &tls_password_callback);
        SSL_CTX_set_default_passwd_cb_userdata(current_settings->tls_context, (void *)current_settings);
        }

      if ((current_settings->tls_certificate_path != NULL) &&
          ((tls_return = SSL_CTX_use_certificate_chain_file(current_settings->tls_context, current_settings->tls_certificate_path)) != 1))
        {
        tls_error(current_settings, tls_return, SYSLOG_ERROR_TLS_CERTIFICATE, current_settings->tls_certificate_path);
        error_occurred = 1;
        }

      if (!error_occurred &&
          (current_settings->tls_privatekey_path != NULL))
        {
        if ((tls_return = SSL_CTX_use_PrivateKey_file(current_settings->tls_context, current_settings->tls_privatekey_path, SSL_FILETYPE_PEM)) != 1)
          {
          tls_error(current_settings, tls_return, SYSLOG_ERROR_TLS_PRIVATEKEY, current_settings->tls_privatekey_path);
          error_occurred = 1;
          }
        }
      else if (!error_occurred &&
               (current_settings->tls_certificate_path != NULL) &&
               ((tls_return = SSL_CTX_use_PrivateKey_file(current_settings->tls_context, current_settings->tls_certificate_path, SSL_FILETYPE_PEM)) != 1))
        {
        tls_error(current_settings, tls_return, SYSLOG_ERROR_TLS_PRIVATEKEY, current_settings->tls_certificate_path);
        error_occurred = 1;
        }

      if (!error_occurred &&
          ((current_settings->tls_session = SSL_new(current_settings->tls_context)) != NULL))
        return_value = 1;
      else
        {
        syslog(LOG_ERR, SYSLOG_ERROR_TLS_INIT);

        SSL_CTX_free(current_settings->tls_context);
        current_settings->tls_context = NULL;
        }
      }
    else
      syslog(LOG_ERR, SYSLOG_ERROR_TLS_INIT);
    }

  return(return_value);
  }

int tls_start(struct filter_settings *current_settings, int read_fd, int write_fd)
  {
  int return_value;
  int tls_return;

  return_value = 0;

  if ((current_settings->tls_session != NULL) &&
      SSL_set_rfd(current_settings->tls_session, read_fd) &&
      SSL_set_wfd(current_settings->tls_session, write_fd))
    if ((tls_return = SSL_accept(current_settings->tls_session)) == 1)
      {
      SSL_set_mode(current_settings->tls_session, SSL_MODE_ENABLE_PARTIAL_WRITE);
      SSL_set_mode(current_settings->tls_session, SSL_MODE_AUTO_RETRY);
      current_settings->tls_state = TLS_STATE_ACTIVE_SPAMDYKE;

      return_value = 1;
      }
    else
      tls_error(current_settings, tls_return, SYSLOG_ERROR_TLS_ACCEPT, NULL);
  else
    syslog(LOG_ERR, SYSLOG_ERROR_TLS_INIT);

  return(return_value);
  }

int tls_end(struct filter_settings *current_settings, int read_fd)
  {
  int return_value;
  fd_set read_fds;
  struct timeval tmp_timeout;

  return_value = 0;

  if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
    {
    if (!SSL_shutdown(current_settings->tls_session))
      {
      /* The socket is going to be closed, so proper SSL shutdown isn't a big
         deal.  We'll give it a chance to happen though. */
      FD_ZERO(&read_fds);
      FD_SET(read_fd, &read_fds);

      tmp_timeout.tv_sec = TIMEOUT_TLS_SHUTDOWN_SECS;
      tmp_timeout.tv_usec = 0;

      select(read_fd + 1, &read_fds, NULL, NULL, &tmp_timeout);

      SSL_shutdown(current_settings->tls_session);
      }

    current_settings->tls_state = TLS_STATE_INACTIVE;
    return_value = 1;
    }

  if (current_settings->tls_session != NULL)
    {
    /* NOTE: SSL_free() may not actually free the SSL object if the reference
       count is above zero.  There's no way to tell if it did, however. */
    SSL_free(current_settings->tls_session);
    current_settings->tls_session = NULL;
    }

  if (current_settings->tls_session != NULL)
    {
    /* NOTE: SSL_CTX_free() may not actually free the SSL object if the
       reference count is above zero.  There's no way to tell if it did,
       however. */
    SSL_CTX_free(current_settings->tls_context);
    current_settings->tls_context = NULL;
    }

  return(return_value);
  }

int tls_can_read(struct filter_settings *current_settings)
  {
  return((current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE) ? SSL_pending(current_settings->tls_session) : 0);
  }

ssize_t tls_read(struct filter_settings *current_settings, int target_fd, void *target_buf, size_t num_bytes)
  {
  ssize_t return_value;

  if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
    return_value = SSL_read(current_settings->tls_session, target_buf, num_bytes);
  else
    return_value = read(target_fd, target_buf, num_bytes);

  return(return_value);
  }

ssize_t tls_write(struct filter_settings *current_settings, int target_fd, void *target_buf, size_t num_bytes)
  {
  ssize_t return_value;

  if (current_settings->tls_state == TLS_STATE_ACTIVE_SPAMDYKE)
    return_value = SSL_write(current_settings->tls_session, target_buf, num_bytes);
  else
    return_value = write(target_fd, target_buf, num_bytes);

  return(return_value);
  }

#endif // TLS
